#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import logging

import progressbar
import xdg

from olstests import fixtures


def isolate_for_config(test):
    """Isolate a test from xdg so a private temp config is used.

    :param test: A test providing a 'uniq_dir' attribute pointing to a private
        removed at the end of the test.
    """
    # xdg use module level global vars that need to point under our private
    # dir.
    fixtures.patch(test, xdg.BaseDirectory, 'xdg_config_home',
                   os.path.join(test.uniq_dir, '.config'))
    fixtures.patch(test, xdg.BaseDirectory, 'xdg_config_dirs',
                   [xdg.BaseDirectory.xdg_config_home])


def use_staging_store(test):
        fixtures.override_env(
            test, 'UBUNTU_SSO_API_ROOT_URL',
            'https://login.staging.ubuntu.com/api/v2/')
        fixtures.override_env(
            test, 'UBUNTU_STORE_API_ROOT_URL',
            'https://myapps.developer.staging.ubuntu.com/dev/api/')
        fixtures.override_env(
            test, 'UBUNTU_STORE_SEARCH_ROOT_URL',
            'https://search.apps.staging.ubuntu.com/')
        fixtures.override_env(
            test, 'UBUNTU_STORE_UPLOAD_ROOT_URL',
            'https://upload.apps.staging.ubuntu.com/')


class SilentProgressBar(progressbar.ProgressBar):
    """A progress bar causing no spurious output during tests."""
    def start(self):
        pass

    def update(self, value=None):
        pass

    def finish(self):
        pass


def override_logging(test, stream_out, level=logging.INFO, fmt='%(message)s'):
    """Setup a logging handler, restoring the actual handlers after the test.

    This assumes a logging setup where handlers are added to the root logger
    only.

    :param stream_out: A stream where log messges will be recorded.

    :param level: The logging level to be active during the test.

    :param fmt: The logging format to use during the test.

    """
    root_logger = logging.getLogger(None)
    # Using reversed() below ensures we can modify root_logger.handlers as well
    # as providing the handlers in the right order for cleanups.
    for handler in reversed(root_logger.handlers):
        root_logger.removeHandler(handler)
        test.addCleanup(root_logger.addHandler, handler)
    # Install the new handler
    new_handler = logging.StreamHandler(stream_out)
    test.addCleanup(root_logger.removeHandler, new_handler)
    root_logger.addHandler(new_handler)
    # Install the new level, restoring the actual one after the test
    test.addCleanup(root_logger.setLevel, root_logger.level)
    root_logger.setLevel(level)
