#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from olstests import scenarii

import olsstore

from olsstore import tests


load_tests = scenarii.load_tests_with_scenarios


class TestDownloadLogin(tests.RecordedTestCase):

    def test_download_without_credentials(self):
        with self.assertRaises(olsstore.InvalidCredentialsError):
            self.download('basic')


class TestSearchPackage(tests.RecordedTestCase):

    def setUp(self):
        super().setUp()
        self.addCleanup(self.logout)
        self.login()
        self.cpi = olsstore.CPIClient(self.store.conf)
        self.addCleanup(self.cpi.close)

    def test_search_known_package(self):
        # We rely on a package that has been published for TEST_USER_EMAIL
        pkg = self.cpi.search_package('femto', 'stable', 'amd64')
        self.assertIn('download_url', pkg)
        self.assertIn('download_sha512', pkg)

    def test_search_unknown_package(self):
        self.assertIsNone(self.cpi.search_package('dont-exist', 'dont-exist',
                                                  'amd64'))


class TestDownload(tests.RecordedTestCase):

    def setUp(self):
        super().setUp()
        self.addCleanup(self.logout)
        self.login()

    def assertLog(self, expected):
        self.assertEqual(expected, self.log_stream.getvalue())

    def test_download_without_login(self):
        self.logout()
        pkg_name = 'femto'
        with self.assertRaises(olsstore.InvalidCredentialsError):
            self.download(pkg_name, 'stable', 'downloaded.snap', 'amd64')

    def test_download_not_found(self):
        with self.assertRaises(olsstore.SnapNotFoundError) as cm:
            self.download('gloo', 'bee', 'downloaded.snap', 'amd64')
        self.assertEqual(
            'The "gloo" for amd64 was not found in bee.',
            str(cm.exception))

    def test_download_mismatch(self):
        self.addCleanup(setattr, olsstore.StoreClient, 'download',
                        olsstore.StoreClient.download)

        def raise_not_a_sha(*args):
            raise olsstore.SHAMismatchError('downloaded.snap', 'not-a-sha')
        olsstore.StoreClient.download = raise_not_a_sha
        with self.assertRaises(olsstore.SHAMismatchError) as cm:
            self.download('femto', 'stable', 'downloaded.snap', 'amd64')
        self.assertEqual(
            'SHA512 checksum for downloaded.snap is not not-a-sha.',
            str(cm.exception))

    def test_download_unknown_package(self):
        with self.assertRaises(olsstore.SnapNotFoundError) as cm:
            self.download('gloo', 'bee')
        self.assertEqual('gloo', cm.exception.name)
        self.assertEqual('bee', cm.exception.channel)
        self.assertEqual('amd64', cm.exception.arch)  # default value

    def test_download_works(self):
        pkg_name = 'femto'
        self.download(pkg_name, 'stable', 'downloaded.snap')
        self.assertTrue(os.path.exists('downloaded.snap'))
        self.assertLog('''Getting details for {pkg_name}
Downloading {pkg_name}
Successfully downloaded {pkg_name} at downloaded.snap
'''.format(pkg_name=pkg_name))

    def test_download_twice_reuse_existing(self):
        pkg_name = 'femto'
        self.download(pkg_name, 'stable', 'downloaded.snap')
        self.assertTrue(os.path.exists('downloaded.snap'))
        self.assertLog('''Getting details for {pkg_name}
Downloading {pkg_name}
Successfully downloaded {pkg_name} at downloaded.snap
'''.format(pkg_name=pkg_name))
        # Try downloading again
        self.download(pkg_name, 'stable', 'downloaded.snap')
        self.assertLog('''Getting details for {pkg_name}
Downloading {pkg_name}
Successfully downloaded {pkg_name} at downloaded.snap
Getting details for {pkg_name}
Already downloaded {pkg_name} at downloaded.snap
'''.format(pkg_name=pkg_name))

    def test_redownload_on_mismatch(self):
        pkg_name = 'femto'
        self.download(pkg_name, 'stable', 'downloaded.snap')
        self.assertTrue(os.path.exists('downloaded.snap'))
        self.assertLog('''Getting details for {pkg_name}
Downloading {pkg_name}
Successfully downloaded {pkg_name} at downloaded.snap
'''.format(pkg_name=pkg_name))
        # Clobber the downloaded file
        with open('downloaded.snap', 'w') as f:
            f.write('Sabotage !')
        # Try downloading again
        self.download(pkg_name, 'stable', 'downloaded.snap')
        self.assertLog('''Getting details for {pkg_name}
Downloading {pkg_name}
Successfully downloaded {pkg_name} at downloaded.snap
Getting details for {pkg_name}
Downloading {pkg_name}
Successfully downloaded {pkg_name} at downloaded.snap
'''.format(pkg_name=pkg_name))
